from ctypes import byref, c_char_p, c_size_t
from pathlib import Path
from typing import Optional, Union

from numpy import empty, ndarray

from expelliarmus.utils import _SUPPORTED_ENCODINGS
from expelliarmus.wizard.clib import (
    c_cargos_t,
    c_cut_fns,
    c_get_time_window_fns,
    c_measure_fns,
    c_read_fns,
    c_save_fns,
    dat_cargo_t,
    event_t,
    events_cargo_t,
    evt2_cargo_t,
    evt3_cargo_t,
)


def c_read_wrapper(encoding: str, fpath: Union[str, Path], buff_size: int):
    c_fpath = c_char_p(bytes(str(fpath), "utf-8"))
    c_buff_size = c_size_t(buff_size)
    cargo = c_cargos_t[encoding](events_info=events_cargo_t())
    c_measure_fns[encoding](c_fpath, byref(cargo), c_buff_size)
    status = 0
    if cargo.events_info.dim > 0:
        arr = empty((cargo.events_info.dim,), dtype=event_t)
        status = c_read_fns[encoding](c_fpath, arr, byref(cargo), c_buff_size)
    return (
        (arr, status) if cargo.events_info.dim > 0 and status == 0 else (None, status)
    )


def c_save_wrapper(
    encoding: str,
    fpath: Union[str, Path],
    arr: ndarray,
    buff_size: int,
):
    c_fpath = c_char_p(bytes(str(fpath), "utf-8"))
    c_buff_size = c_size_t(buff_size)
    dim = len(arr)
    if arr.dtype != event_t:
        loc_arr = empty((dim,), dtype=event_t)
        for k in ("t", "y", "x", "p"):
            loc_arr[k] = arr[k].astype(loc_arr[k].dtype)
    else:
        loc_arr = arr
    cargo = c_cargos_t[encoding](events_info=events_cargo_t(dim=dim, start_byte=0))
    return c_save_fns[encoding](c_fpath, loc_arr, byref(cargo), c_buff_size)


def c_read_time_window_wrapper(
    encoding: str,
    fpath: Union[str, Path],
    cargo: Union[dat_cargo_t, evt2_cargo_t, evt3_cargo_t],
    buff_size: int,
):
    c_fpath = c_char_p(bytes(str(fpath), "utf-8"))
    c_buff_size = c_size_t(buff_size)

    c_get_time_window_fns[encoding](c_fpath, byref(cargo), c_buff_size)
    status = 0
    if cargo.events_info.dim > 0:
        arr = empty((cargo.events_info.dim,), dtype=event_t)
        status = c_read_fns[encoding](c_fpath, arr, byref(cargo), c_buff_size)
    return (
        (arr, cargo, status)
        if cargo.events_info.dim > 0 and status == 0
        else (None, cargo, status)
    )


def c_read_chunk_wrapper(
    encoding: str,
    fpath: Union[str, Path],
    cargo: Union[dat_cargo_t, evt2_cargo_t, evt3_cargo_t],
    buff_size: int,
):
    c_fpath = c_char_p(bytes(str(fpath), "utf-8"))
    c_buff_size = c_size_t(buff_size)
    arr = empty(
        (cargo.events_info.dim + (12 if encoding == "evt3" else 0),), dtype=event_t
    )
    status = c_read_fns[encoding](c_fpath, arr, byref(cargo), c_buff_size)
    return (
        (arr[: cargo.events_info.dim], cargo, status)
        if cargo.events_info.dim > 0 and status == 0
        else (None, cargo, status)
    )


def c_cut_wrapper(
    encoding: str,
    fpath_in: Union[str, Path],
    fpath_out: Union[str, Path],
    new_duration: int,
    buff_size: int,
):
    c_fpath_in = c_char_p(bytes(str(fpath_in), "utf-8"))
    c_fpath_out = c_char_p(bytes(str(fpath_out), "utf-8"))
    c_new_duration = c_size_t(new_duration)
    c_buff_size = c_size_t(buff_size)
    return c_cut_fns[encoding](c_fpath_in, c_fpath_out, c_new_duration, c_buff_size)



import os
import pathlib
import re
from ctypes import (
    CDLL,
    POINTER,
    Structure,
    c_char_p,
    c_int,
    c_int16,
    c_int64,
    c_size_t,
    c_uint,
    c_uint8,
    c_uint16,
    c_uint64,
)

from numpy import zeros
from numpy.ctypeslib import ndpointer

from expelliarmus.utils import _ROOT_PATH

# Searching for the shared library.
lib_re = r"^expelliarmus\..*\.(so|pyd)$"
for root, dirs, files in os.walk(_ROOT_PATH):
    for f in files:
        if re.match(lib_re, f):
            lib_path = pathlib.Path(os.path.join(root, f))
            break
clib = CDLL(str(lib_path))


class event_t(Structure):
    _fields_ = [
        ("t", c_int64),
        ("x", c_int16),
        ("y", c_int16),
        ("p", c_uint8),
    ]


class events_cargo_t(Structure):
    _fields_ = [
        ("dim", c_size_t),
        ("is_chunk", c_uint8),
        ("time_window", c_size_t),
        ("is_time_window", c_uint8),
        ("start_byte", c_size_t),
        ("finished", c_uint8),
    ]


class dat_cargo_t(Structure):
    _fields_ = [
        ("events_info", events_cargo_t),
        ("last_t", c_uint64),
        ("time_ovfs", c_uint64),
    ]


class evt2_cargo_t(Structure):
    _fields_ = [
        ("events_info", events_cargo_t),
        ("last_t", c_int64),
        ("time_high", c_uint64),
    ]


class evt3_cargo_t(Structure):
    _fields_ = [
        ("events_info", events_cargo_t),
        ("time_high", c_uint64),
        ("time_low", c_uint64),
        ("time_high_ovfs", c_uint64),
        ("time_low_ovfs", c_uint64),
        ("base_x", c_uint16),
        ("last_event", event_t),
    ]


c_cargos_t = dict(dat=dat_cargo_t, evt2=evt2_cargo_t, evt3=evt3_cargo_t)

# Setting up C wrappers.
# Read functions.
c_read_dat = clib.read_dat
c_read_evt2 = clib.read_evt2
c_read_evt3 = clib.read_evt3

for fn, cargo_t in zip(
    (c_read_dat, c_read_evt2, c_read_evt3), (dat_cargo_t, evt2_cargo_t, evt3_cargo_t)
):
    fn.argtypes = [
        c_char_p,
        ndpointer(ndim=1),
        POINTER(cargo_t),
        c_size_t,
    ]
    fn.restype = c_int

c_read_fns = dict(dat=c_read_dat, evt2=c_read_evt2, evt3=c_read_evt3)

# Compression functions.
c_save_dat = clib.save_dat
c_save_evt2 = clib.save_evt2
c_save_evt3 = clib.save_evt3

for fn, cargo_t in zip(
    (c_save_dat, c_save_evt2, c_save_evt3), (dat_cargo_t, evt2_cargo_t, evt3_cargo_t)
):
    fn.argtypes = [
        c_char_p,
        ndpointer(ndim=1),
        POINTER(cargo_t),
        c_size_t,
    ]
    fn.restype = c_int

c_save_fns = dict(dat=c_save_dat, evt2=c_save_evt2, evt3=c_save_evt3)

# Cut functions.
ARGTYPES_CUT = [c_char_p, c_char_p, c_size_t, c_size_t]
RESTYPE_CUT = c_size_t

c_cut_dat = clib.cut_dat
c_cut_evt2 = clib.cut_evt2
c_cut_evt3 = clib.cut_evt3
for fn in (c_cut_dat, c_cut_evt2, c_cut_evt3):
    fn.restype = RESTYPE_CUT
    fn.argtypes = ARGTYPES_CUT

c_cut_fns = dict(dat=c_cut_dat, evt2=c_cut_evt2, evt3=c_cut_evt3)

# Measure functions.
c_measure_dat = clib.measure_dat
c_measure_evt2 = clib.measure_evt2
c_measure_evt3 = clib.measure_evt3
for fn, cargo_t in zip(
    (c_measure_dat, c_measure_evt2, c_measure_evt3),
    (dat_cargo_t, evt2_cargo_t, evt3_cargo_t),
):
    fn.argtypes = [c_char_p, POINTER(cargo_t), c_size_t]
    fn.restype = None

c_measure_fns = dict(dat=c_measure_dat, evt2=c_measure_evt2, evt3=c_measure_evt3)

# Time window functions.
c_get_time_window_dat = clib.get_time_window_dat
c_get_time_window_evt2 = clib.get_time_window_evt2
c_get_time_window_evt3 = clib.get_time_window_evt3
for fn, cargo_t in zip(
    (c_get_time_window_dat, c_get_time_window_evt2, c_get_time_window_evt3),
    (dat_cargo_t, evt2_cargo_t, evt3_cargo_t),
):
    fn.argtypes = [c_char_p, POINTER(cargo_t), c_size_t]
    fn.restype = None

c_get_time_window_fns = dict(
    dat=c_get_time_window_dat, evt2=c_get_time_window_evt2, evt3=c_get_time_window_evt3
)
